@@ -16,11 +16,15 @@
 
 import io
 import json
+import os
 from pickle import PicklingError
 
 import six
 
+import google.api_core.client_options
+import google.api_core.exceptions
 import google.auth
+from google.auth import environment_vars
 import google.auth.credentials
 import google.auth.transport.requests
 from google.cloud._helpers import _determine_default_project
@@ -47,6 +51,36 @@ class _ClientFactoryMixin(object):
 
     _SET_PROJECT = False
 
+    @classmethod
+    def from_service_account_info(cls, info, *args, **kwargs):
+        """Factory to retrieve JSON credentials while creating client.
+
+        :type info: str
+        :param info:
+            The JSON object with a private key and other credentials
+            information (downloaded from the Google APIs console).
+
+        :type args: tuple
+        :param args: Remaining positional arguments to pass to constructor.
+
+        :param kwargs: Remaining keyword arguments to pass to constructor.
+
+        :rtype: :class:`_ClientFactoryMixin`
+        :returns: The client created with the retrieved JSON credentials.
+        :raises TypeError: if there is a conflict with the kwargs
+                 and the credentials created by the factory.
+        """
+        if "credentials" in kwargs:
+            raise TypeError("credentials must not be in keyword arguments")
+
+        credentials = service_account.Credentials.from_service_account_info(info)
+        if cls._SET_PROJECT:
+            if "project" not in kwargs:
+                kwargs["project"] = info.get("project_id")
+
+        kwargs["credentials"] = credentials
+        return cls(*args, **kwargs)
+
     @classmethod
     def from_service_account_json(cls, json_credentials_path, *args, **kwargs):
         """Factory to retrieve JSON credentials while creating client.
@@ -69,19 +103,10 @@ class _ClientFactoryMixin(object):
         :raises TypeError: if there is a conflict with the kwargs
                  and the credentials created by the factory.
         """
-        if "credentials" in kwargs:
-            raise TypeError("credentials must not be in keyword arguments")
         with io.open(json_credentials_path, "r", encoding="utf-8") as json_fi:
             credentials_info = json.load(json_fi)
-        credentials = service_account.Credentials.from_service_account_info(
-            credentials_info
-        )
-        if cls._SET_PROJECT:
-            if "project" not in kwargs:
-                kwargs["project"] = credentials_info.get("project_id")
 
-        kwargs["credentials"] = credentials
-        return cls(*args, **kwargs)
+        return cls.from_service_account_info(credentials_info)
 
 
 class Client(_ClientFactoryMixin):
@@ -102,6 +127,8 @@ class Client(_ClientFactoryMixin):
             (Optional) The OAuth2 Credentials to use for this client. If not
             passed (and if no ``_http`` object is passed), falls back to the
             default inferred from the environment.
+        client_options (google.api_core.client_options.ClientOptions):
+            (Optional) Custom options for the client.
         _http (requests.Session):
             (Optional) HTTP object to make requests. Can be any object that
             defines ``request()`` with the same interface as
@@ -123,17 +150,44 @@ class Client(_ClientFactoryMixin):
     Needs to be set by subclasses.
     """
 
-    def __init__(self, credentials=None, _http=None):
-        if credentials is not None and not isinstance(
+    def __init__(self, credentials=None, _http=None, client_options=None):
+        if isinstance(client_options, dict):
+            client_options = google.api_core.client_options.from_dict(client_options)
+        if client_options is None:
+            client_options = google.api_core.client_options.ClientOptions()
+
+        if credentials and client_options.credentials_file:
+            raise google.api_core.exceptions.DuplicateCredentialArgs(
+                "'credentials' and 'client_options.credentials_file' are mutually exclusive."
+            )
+
+        if credentials and not isinstance(
             credentials, google.auth.credentials.Credentials
         ):
             raise ValueError(_GOOGLE_AUTH_CREDENTIALS_HELP)
-        if credentials is None and _http is None:
-            credentials, _ = google.auth.default()
+
+        scopes = client_options.scopes or self.SCOPE
+
+        # if no http is provided, credentials must exist
+        if not _http and credentials is None:
+            if client_options.credentials_file:
+                credentials, _ = google.auth.load_credentials_from_file(
+                    client_options.credentials_file, scopes=scopes
+                )
+            else:
+                credentials, _ = google.auth.default(scopes=scopes)
+
         self._credentials = google.auth.credentials.with_scopes_if_required(
-            credentials, self.SCOPE
+            credentials, scopes=scopes
         )
+
+        if client_options.quota_project_id:
+            self._credentials = self._credentials.with_quota_project(
+                client_options.quota_project_id
+            )
+
         self._http_internal = _http
+        self._client_cert_source = client_options.client_cert_source
 
     def __getstate__(self):
         """Explicitly state that clients are not pickleable."""
@@ -155,36 +209,74 @@ class Client(_ClientFactoryMixin):
         """
         if self._http_internal is None:
             self._http_internal = google.auth.transport.requests.AuthorizedSession(
-                self._credentials,
-                refresh_timeout=_CREDENTIALS_REFRESH_TIMEOUT,
+                self._credentials, refresh_timeout=_CREDENTIALS_REFRESH_TIMEOUT,
             )
+            self._http_internal.configure_mtls_channel(self._client_cert_source)
         return self._http_internal
 
+    def close(self):
+        """Clean up transport, if set.
+
+        Suggested use:
+
+        .. code-block:: python
+
+           import contextlib
+
+           with contextlib.closing(client):  # closes on exit
+               do_something_with(client)
+        """
+        if self._http_internal is not None:
+            self._http_internal.close()
+
 
 class _ClientProjectMixin(object):
     """Mixin to allow setting the project on the client.
 
     :type project: str
-    :param project: the project which the client acts on behalf of. If not
-                    passed falls back to the default inferred from the
-                    environment.
+    :param project:
+        (Optional) the project which the client acts on behalf of. If not
+        passed, falls back to the default inferred from the environment.
+
+    :type credentials: :class:`google.auth.credentials.Credentials`
+    :param credentials:
+        (Optional) credentials used to discover a project, if not passed.
 
     :raises: :class:`EnvironmentError` if the project is neither passed in nor
-             set in the environment. :class:`ValueError` if the project value
-             is invalid.
+             set on the credentials or in the environment. :class:`ValueError`
+             if the project value is invalid.
     """
 
-    def __init__(self, project=None):
-        project = self._determine_default(project)
+    def __init__(self, project=None, credentials=None):
+        # This test duplicates the one from `google.auth.default`, but earlier,
+        # for backward compatibility:  we want the environment variable to
+        # override any project set on the credentials.  See:
+        # https://github.com/googleapis/python-cloud-core/issues/27
+        if project is None:
+            project = os.getenv(
+                environment_vars.PROJECT, os.getenv(environment_vars.LEGACY_PROJECT),
+            )
+
+        # Project set on explicit credentials overrides discovery from
+        # SDK / GAE / GCE.
+        if project is None and credentials is not None:
+            project = getattr(credentials, "project_id", None)
+
+        if project is None:
+            project = self._determine_default(project)
+
         if project is None:
             raise EnvironmentError(
                 "Project was not passed and could not be "
                 "determined from the environment."
             )
+
         if isinstance(project, six.binary_type):
             project = project.decode("utf-8")
+
         if not isinstance(project, six.string_types):
             raise ValueError("Project must be a string.")
+
         self.project = project
 
     @staticmethod
@@ -222,6 +314,8 @@ class ClientWithProject(Client, _ClientProjectMixin):
 
     _SET_PROJECT = True  # Used by from_service_account_json()
 
-    def __init__(self, project=None, credentials=None, _http=None):
-        _ClientProjectMixin.__init__(self, project=project)
-        Client.__init__(self, credentials=credentials, _http=_http)
+    def __init__(self, project=None, credentials=None, client_options=None, _http=None):
+        _ClientProjectMixin.__init__(self, project=project, credentials=credentials)
+        Client.__init__(
+            self, credentials=credentials, client_options=client_options, _http=_http
+        )
